<?php

namespace SGalinski\SgJobs\Property\TypeConverter;

/***************************************************************
 *  Copyright notice
 *
 *  (c) 2014 Helmut Hummel
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use TYPO3\CMS\Core\Resource\DuplicationBehavior;
use TYPO3\CMS\Core\Resource\File as FalFile;
use TYPO3\CMS\Core\Resource\FileReference as FalFileReference;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\PathUtility;
use TYPO3\CMS\Extbase\Domain\Model\FileReference;
use TYPO3\CMS\Extbase\Error\Error;
use TYPO3\CMS\Extbase\Property\Exception\TypeConverterException;
use TYPO3\CMS\Extbase\Property\PropertyMappingConfigurationInterface;
use TYPO3\CMS\Extbase\Property\TypeConverter\AbstractTypeConverter;

/**
 * Class UploadedFileReferenceConverter
 */
class UploadedFileReferenceConverter extends AbstractTypeConverter {

	/**
	 * Folder where the file upload should go to (including storage).
	 *
	 * @var string
	 */
	protected $uploadFolder = '1:/user_upload/';

	/**
	 * How to handle a upload when the name of the uploaded file conflicts.
	 *
	 * @var string
	 */
	protected $uploadConflictMode = DuplicationBehavior::RENAME;

	/**
	 * Whether to replace an already present resource.
	 * Useful for "maxitems = 1" fields and properties
	 * with no ObjectStorage annotation.
	 *
	 * @var string
	 */
	protected $allowedFileExtensions = '';

	/**
	 * @var array<string>
	 */
	protected $sourceTypes = ['array'];

	/**
	 * @var string
	 */
	protected $targetType = FileReference::class;

	/**
	 * Take precedence over the available FileReferenceConverter
	 *
	 * @var int
	 */
	protected $priority = 30;

	/**
	 * @var \TYPO3\CMS\Core\Resource\ResourceFactory
	 * @inject
	 */
	protected $resourceFactory;

	/**
	 * @var \TYPO3\CMS\Extbase\Security\Cryptography\HashService
	 * @inject
	 */
	protected $hashService;

	/**
	 * @var \TYPO3\CMS\Extbase\Persistence\Generic\PersistenceManager
	 * @inject
	 */
	protected $persistenceManager;

	/**
	 * @var \TYPO3\CMS\Core\Resource\FileInterface[]
	 */
	protected $convertedResources = [];

	/**
	 * @var string
	 */
	protected $targetUploadFileName = '';

	/**
	 * @return string
	 */
	public function getTargetUploadFileName(): string {
		return $this->targetUploadFileName;
	}

	/**
	 * @param string $targetUploadFileName
	 * @return void
	 */
	public function setTargetUploadFileName($targetUploadFileName) {
		$this->targetUploadFileName = $targetUploadFileName;
	}

	/**
	 * @return string
	 */
	public function getUploadFolder(): string {
		return $this->uploadFolder;
	}

	/**
	 * @param string $uploadFolder
	 * @return void
	 */
	public function setUploadFolder($uploadFolder) {
		$this->uploadFolder = $uploadFolder;
	}

	/**
	 * @return string
	 */
	public function getUploadConflictMode(): string {
		return $this->uploadConflictMode;
	}

	/**
	 * @param string $uploadConflictMode
	 * @return void
	 */
	public function setUploadConflictMode($uploadConflictMode) {
		$this->uploadConflictMode = $uploadConflictMode;
	}

	/**
	 * @return string
	 */
	public function getAllowedFileExtensions(): string {
		return $this->allowedFileExtensions;
	}

	/**
	 * @param string $allowedFileExtensions
	 * @return void
	 */
	public function setAllowedFileExtensions($allowedFileExtensions) {
		$this->allowedFileExtensions = $allowedFileExtensions;
	}

	/**
	 * Actually convert from $source to $targetType, taking into account the fully
	 * built $convertedChildProperties and $configuration.
	 *
	 * @param array $source
	 * @param string $targetType
	 * @param array $convertedChildProperties
	 * @param \TYPO3\CMS\Extbase\Property\PropertyMappingConfigurationInterface $configuration
	 * @return null|\TYPO3\CMS\Core\Resource\FileInterface|FileReference|Error
	 * @throws \TYPO3\CMS\Extbase\Security\Exception\InvalidHashException
	 * @throws \TYPO3\CMS\Extbase\Security\Exception\InvalidArgumentForHashGenerationException
	 * @throws \TYPO3\CMS\Core\Resource\Exception\FileDoesNotExistException
	 * @api
	 * @throws \TYPO3\CMS\Core\Resource\Exception\ResourceDoesNotExistException
	 */
	public function convertFrom(
		$source, $targetType, array $convertedChildProperties = [],
		PropertyMappingConfigurationInterface $configuration = NULL
	) {
		if (!isset($source['error']) || $source['error'] === \UPLOAD_ERR_NO_FILE) {
			if (isset($source['submittedFile']['resourcePointer'])) {
				try {
					$resourcePointer = $this->hashService->validateAndStripHmac(
						$source['submittedFile']['resourcePointer']
					);
					if (strpos($resourcePointer, 'file:') === 0) {
						$fileUid = substr($resourcePointer, 5);
						return $this->createFileReferenceFromFalFileObject(
							$this->resourceFactory->getFileObject($fileUid)
						);
					}

					return $this->createFileReferenceFromFalFileReferenceObject(
						$this->resourceFactory->getFileReferenceObject($resourcePointer), $resourcePointer
					);
				} catch (\InvalidArgumentException $e) {
					// Nothing to do. No file is uploaded and resource pointer is invalid. Discard!
				}
			}
			return NULL;
		}

		if ($source['error'] !== \UPLOAD_ERR_OK) {
			switch ($source['error']) {
				case \UPLOAD_ERR_INI_SIZE:
				case \UPLOAD_ERR_FORM_SIZE:
				case \UPLOAD_ERR_PARTIAL:
					return new Error('Error Code: ' . $source['error'], 1264440823);
				default:
					return new Error(
						'An error occurred while uploading. Please try again or contact the administrator if the problem remains',
						1340193849
					);
			}
		}

		if (isset($this->convertedResources[$source['tmp_name']])) {
			return $this->convertedResources[$source['tmp_name']];
		}

		try {
			if ($configuration) {
				$resource = $this->importUploadedResource($source, $configuration);
			}
		} catch (\Exception $e) {
			return new Error($e->getMessage(), $e->getCode());
		}

		$this->convertedResources[$source['tmp_name']] = $resource;
		return $resource;
	}

	/**
	 * Import a resource and respect configuration given for properties
	 *
	 * @param array $uploadInfo
	 * @param PropertyMappingConfigurationInterface $configuration
	 * @return \SGalinski\SgJobs\Domain\Model\FileReference
	 * @throws TypeConverterException
	 * @throws \TYPO3\CMS\Extbase\Security\Exception\InvalidArgumentForHashGenerationException
	 * @throws \TYPO3\CMS\Extbase\Security\Exception\InvalidHashException
	 */
	protected function importUploadedResource(array $uploadInfo, PropertyMappingConfigurationInterface $configuration
	): FileReference {
		if (!GeneralUtility::verifyFilenameAgainstDenyPattern($uploadInfo['name'])) {
			throw new TypeConverterException('Uploading files with PHP file extensions is not allowed!', 1399312430);
		}

		$allowedFileExtensions = $this->getAllowedFileExtensions();
		$filePathInfo = PathUtility::pathinfo($uploadInfo['name']);
		if ($allowedFileExtensions !== NULL) {
			if (!GeneralUtility::inList($allowedFileExtensions, strtolower($filePathInfo['extension']))) {
				throw new TypeConverterException('File extension is not allowed!', 1399312430);
			}
		}

		$uploadFolderId = $this->getUploadFolder();
		$conflictMode = $this->getUploadConflictMode();
		$uploadFolder = $this->resourceFactory->retrieveFileOrFolderObject($uploadFolderId);
		$uploadedFile = $uploadFolder->addUploadedFile($uploadInfo, $conflictMode);
		$uploadedFile->rename($this->getTargetUploadFileName() . '.' . $filePathInfo['extension']);

		$resourcePointer = isset($uploadInfo['submittedFile']['resourcePointer']) &&
		strpos($uploadInfo['submittedFile']['resourcePointer'], 'file:') === FALSE
			? $this->hashService->validateAndStripHmac($uploadInfo['submittedFile']['resourcePointer'])
			: NULL;

		$fileReferenceModel = $this->createFileReferenceFromFalFileObject($uploadedFile, $resourcePointer);

		return $fileReferenceModel;
	}

	/**
	 * @param FalFile $file
	 * @param int $resourcePointer
	 * @return \SGalinski\SgJobs\Domain\Model\FileReference
	 */
	protected function createFileReferenceFromFalFileObject(FalFile $file, $resourcePointer = NULL
	): \SGalinski\SgJobs\Domain\Model\FileReference {
		$fileReference = $this->resourceFactory->createFileReferenceObject(
			[
				'uid_local' => $file->getUid(),
				'uid_foreign' => uniqid('NEW_', TRUE),
				'uid' => uniqid('NEW_', TRUE),
				'crop' => NULL,
			]
		);
		return $this->createFileReferenceFromFalFileReferenceObject($fileReference, $resourcePointer);
	}

	/**
	 * @param FalFileReference $falFileReference
	 * @param int $resourcePointer
	 * @return \SGalinski\SgJobs\Domain\Model\FileReference
	 */
	protected function createFileReferenceFromFalFileReferenceObject(
		FalFileReference $falFileReference, $resourcePointer = NULL
	) {
		if ($resourcePointer === NULL) {
			/** @var \SGalinski\SgJobs\Domain\Model\FileReference $fileReference */
			$fileReference = $this->objectManager->get(FileReference::class);
		} else {
			$fileReference = $this->persistenceManager->getObjectByIdentifier(
				$resourcePointer, FileReference::class, FALSE
			);
		}

		$fileReference->setOriginalResource($falFileReference);

		return $fileReference;
	}
}
